@@ -211,60 +211,114 @@ unsafe fn lookup<T: LoadPage, K: UnsizedStorable + ?Sized, V: UnsizedStorable +
     c.total = hdr.n() as usize;
     c.is_leaf = hdr.is_leaf();
     if c.is_leaf {
-        let s = core::slice::from_raw_parts(
-            page.data.as_ptr().add(HDR) as *const u16,
-            hdr.n() as usize,
-        );
-        if let Some(v0) = v0 {
-            s.binary_search_by(|&off| {
-                let off = u16::from_le(off);
-                let (k, v) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
-                let k = K::from_raw_ptr(txn, k as *const u8);
-                match k.compare(txn, k0) {
-                    Ordering::Equal => {
-                        let v = V::from_raw_ptr(txn, v as *const u8);
-                        v.compare(txn, v0)
-                    }
-                    cmp => cmp,
-                }
-            })
-        } else {
-            s.binary_search_by(|&off| {
-                let off = u16::from_le(off);
-                let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
-                let k = K::from_raw_ptr(txn, k);
-                k.compare(txn, k0)
-            })
-        }
+        lookup_leaf(txn, page, k0, v0, hdr)
     } else {
-        let s = core::slice::from_raw_parts(
-            page.data.as_ptr().add(HDR) as *const u64,
-            hdr.n() as usize,
-        );
-        if let Some(v0) = v0 {
-            s.binary_search_by(|&off| {
-                let off = u64::from_le(off) & 0xfff;
-                let (k, v) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize & 0xfff));
-                let k = K::from_raw_ptr(txn, k);
-                match k.compare(txn, k0) {
-                    Ordering::Equal => {
-                        let v = V::from_raw_ptr(txn, v);
-                        v.compare(txn, v0)
-                    }
-                    cmp => cmp,
+        lookup_internal(txn, page, k0, v0, hdr)
+    }
+}
+
+unsafe fn lookup_internal<T: LoadPage, K: UnsizedStorable + ?Sized, V: UnsizedStorable + ?Sized>(
+    txn: &T,
+    page: crate::Page,
+    k0: &K,
+    v0: Option<&V>,
+    hdr: &header::Header,
+) -> Result<usize, usize> {
+    let s = core::slice::from_raw_parts(
+        page.data.as_ptr().add(HDR) as *const u64,
+        hdr.n() as usize,
+    );
+    if let Some(v0) = v0 {
+        s.binary_search_by(|&off| {
+            let off = u64::from_le(off) & 0xfff;
+            let (k, v) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize & 0xfff));
+            let k = K::from_raw_ptr(txn, k);
+            match k.compare(txn, k0) {
+                Ordering::Equal => {
+                    let v = V::from_raw_ptr(txn, v);
+                    v.compare(txn, v0)
                 }
-            })
-        } else {
-            s.binary_search_by(|&off| {
-                let off = u64::from_le(off) & 0xfff;
-                let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize & 0xfff));
-                let k = K::from_raw_ptr(txn, k);
-                k.compare(txn, k0)
-            })
+                cmp => cmp,
+            }
+        })
+    } else {
+        match s.binary_search_by(|&off| {
+            let off = u64::from_le(off) & 0xfff;
+            let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize & 0xfff));
+            let k = K::from_raw_ptr(txn, k);
+            k.compare(txn, k0)
+        }) {
+            Err(i) => Err(i),
+            Ok(mut i) => {
+                // Rewind if there are multiple matching keys.
+                while i > 0 {
+                    let off = u64::from_le(s[i-1]) & 0xfff;
+                    let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
+                    let k = K::from_raw_ptr(txn, k);
+                    if let Ordering::Equal = k.compare(txn, k0) {
+                        i -= 1
+                    } else {
+                        break
+                    }
+                }
+                Ok(i)
+            }
         }
     }
 }
 
+
+unsafe fn lookup_leaf<T: LoadPage, K: UnsizedStorable + ?Sized, V: UnsizedStorable + ?Sized>(
+    txn: &T,
+    page: crate::Page,
+    k0: &K,
+    v0: Option<&V>,
+    hdr: &header::Header,
+) -> Result<usize, usize> {
+    let s = core::slice::from_raw_parts(
+        page.data.as_ptr().add(HDR) as *const u16,
+        hdr.n() as usize,
+    );
+    if let Some(v0) = v0 {
+        s.binary_search_by(|&off| {
+            let off = u16::from_le(off);
+            let (k, v) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
+            let k = K::from_raw_ptr(txn, k as *const u8);
+            match k.compare(txn, k0) {
+                Ordering::Equal => {
+                    let v = V::from_raw_ptr(txn, v as *const u8);
+                    v.compare(txn, v0)
+                }
+                cmp => cmp,
+            }
+        })
+    } else {
+        match s.binary_search_by(|&off| {
+            let off = u16::from_le(off);
+            let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
+            let k = K::from_raw_ptr(txn, k);
+            k.compare(txn, k0)
+        }) {
+            Err(e) => Err(e),
+            Ok(mut i) => {
+                // Rewind if there are multiple matching keys.
+                while i > 0 {
+                    let off = u16::from_le(s[i-1]);
+                    let (k, _) = read::<T, K, V>(txn, page.data.as_ptr().offset(off as isize));
+                    let k = K::from_raw_ptr(txn, k);
+                    if let Ordering::Equal = k.compare(txn, k0) {
+                        i -= 1
+                    } else {
+                        break
+                    }
+                }
+                Ok(i)
+            }
+        }
+    }
+}
+
+
 impl<
         K: UnsizedStorable + ?Sized + core::fmt::Debug,
         V: UnsizedStorable + ?Sized + core::fmt::Debug,
